/*!
 * @brief vision process module
 * $Date$
 * @author Nemoto,Mizoguchi lab,Tokyo University of Science
 * @mail:j7507640@ed.noda.tus.ac.jp
 * $Id$
 */

#include <stdio.h>
#include <math.h>
#include <cv.h>
#include <highgui.h>
#include <ctype.h>
#include <iostream>
#include <string>
#include "stframe.h"
using namespace std;

void dimensionless(double histgram[][BIT8])
{
    int i,j;
    double center=0;
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    if(center<histgram[i][j])
		center=histgram[i][j];
	}
    }
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    histgram[i][j]=histgram[i][j]/center; 
	}
    }
}



void inputClusterData(IplImage* src,CLdata *cluster){

    for (int y = 0; y < src->height; y++) {
	for (int x = 0; x < src->width; x++) {
	    src->imageData[src->widthStep * y + x * 3] = 0;
	    src->imageData[src->widthStep * y + x * 3 + 1] = 0;
	    src->imageData[src->widthStep * y + x * 3 + 2] = 0;
	}
    }

    for(int j=0;j<cluster->pt;j++){
	if(cluster->clnum[j]==0){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=(char)255;
	}

	else if(cluster->clnum[j]==1){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=0;
	}

	else if(cluster->clnum[j]==2){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=0;
	}

	else if(cluster->clnum[j]==3){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=0;
	}

	else if(cluster->clnum[j]==4){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=(char)255;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=(char)255;
	}
	else if(cluster->clnum[j]==5){
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+1]=0;
	    src->imageData[(cluster->pty[j]*256+cluster->ptx[j])*3+2]=(char)255;
	}
    }
    for(int j=0;j<cluster->clpt;j++){
	src->imageData[(cluster->cly[j]*256+cluster->clx[j])*3]=0;
	src->imageData[(cluster->cly[j]*256+cluster->clx[j])*3+1]=0;
	src->imageData[(cluster->cly[j]*256+cluster->clx[j])*3+2]=(char)255;
    } 
}

void ColorSpaceLoad(HScolor hs[][BIT8][BIT8],string c_space){
    int rr,gg,bb;
    HSVcolor hsv_color;
    double HH,SS,VV;

    for(rr=0;rr<256;rr++){
	fprintf(stderr,"Color Space loading...  %02d%%\r",(int)100*rr/256);
	fflush(stderr);
	for(gg=0;gg<256;gg++){
	    for(bb=0;bb<256;bb++){
		if(c_space=="YCC"){	
		    rgb_YCbCr(rr,gg,bb,&hsv_color);
		}
		else if(c_space=="HSV"){
		    rgb_hsv(rr,gg,bb,&hsv_color);
		}
		HH=hsv_color.H;
		SS=hsv_color.S;
		VV=hsv_color.V;
		hs[rr][gg][bb].H=(int)hsv_color.H;
		hs[rr][gg][bb].S=(int)hsv_color.S;
	    }
	}
    }

    fprintf(stderr,"Color Space loading... 100%%\n");
}

double min_col(unsigned char r, unsigned char g, unsigned char b)
{
    double ret;
    if(r<g){
	if(r<b)
	    ret=r;
	else
	    ret=b;
    }
    else{
	if(g<b)
	    ret=g;
	else
	    ret=b;
    }
    return ret;
}
double max_col(unsigned char r, unsigned char g, unsigned char b)
{
    double ret;
    if(r>g){
	if(r>b)
	    ret=r;
	else
	    ret=b;
    }
    else{
	if(g>b)
	    ret=g;
	else
	    ret=b;
    }
    return ret;
}

	  
void rgb_YCbCr(unsigned char RR, unsigned char GG, unsigned char BB, HSVcolor* YCbCr_color)
{
    double r,b,g;
    double y,cr,cb;
    r=RR/256.0;
    b=BB/256.0;
    g=GG/256.0;
    y=0.299*r+0.587*g+0.114*b;
    cr=0.5*r-0.419*g-0.081*b;
    cb=-0.169*r-0.332*g+0.5*b;
    cr*=256;
    cb*=256;
    cr+=128;
    cb+=128;
    if (y<0)
	y=0;
    if (y>255)
	y=255;
    if (cr<0)
	cr=0;
    if (cr>255)
	cr=255;

    if (cb<0)
	cb=0;
    if (cb>255)
	cb=255;
    YCbCr_color->H= cr;
    YCbCr_color->S= cb;
    YCbCr_color->V= y;
}

void rgb_hsv(unsigned char RR, unsigned char GG, unsigned char BB, HSVcolor* hsvcolor)
{
    double P;
    double R,G,B;

    hsvcolor->V=max_col(RR,GG,BB);
    hsvcolor->V/=255;
    P=min_col(RR,GG,BB);
    P=P/255;

    R=(double)RR/255;
    G=(double)GG/255;
    B=(double)BB/255;
    if(R!=G || G!=B || B!=R){ 
	if(hsvcolor->V==R){
	    hsvcolor->H=60*(G-B)/(hsvcolor->V-P);
	    hsvcolor->S=(hsvcolor->V-P)/(hsvcolor->V);
	}
	else if(hsvcolor->V==G){
	    hsvcolor->H=120+60*(B-R)/(hsvcolor->V-P);
	    hsvcolor->S=(hsvcolor->V-P)/(hsvcolor->V);
	}
	else{
	    hsvcolor->H=HEIGHT+(R-G)*60/(hsvcolor->V-P);
	    hsvcolor->S=(hsvcolor->V-P)/(hsvcolor->V);
	}
    }
    else
	hsvcolor->H=hsvcolor->S=hsvcolor->V=0;

    hsvcolor->H= hsvcolor->H/360;

    hsvcolor->H= hsvcolor->H*255;
    hsvcolor->S= hsvcolor->S*255;
    hsvcolor->V= hsvcolor->V*255;

}

void getGaussian(CLdata* cluster,double histgram[][BIT8], double gaussian[][BIT8])
{
    int i,j,k,n;
    double center=0;
    double prehist[BIT8][BIT8]={{}};
    double f;
    double q[DEG+1]={};
    double F[DEG+1]={};
    int qq;
    double Qtan;
    double dis;
    double fmax=0;
    double max=0;
    int shita=0;
    int Nf=255;

    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    if(center<histgram[i][j])
		center=histgram[i][j];
	}
    }

    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    histgram[i][j]=histgram[i][j]/center;
	}
    }

    for(n=0;n<cluster->clpt;n++){
	max=0;
	for(i=0;i<=cluster->pt;i++){
	    if(cluster->clnum[i]==n){
		if(histgram[cluster->pty[i]][cluster->ptx[i]]>max){
		    max=histgram[cluster->pty[i]][cluster->ptx[i]];
		    cluster->cly[n]=cluster->pty[i];
		    cluster->clx[n]=cluster->ptx[i];
		}
	    }
	}

	for(i=0;i<=cluster->pt;i++){
	    if(cluster->clnum[i]==n){
		if((cluster->pty[i]!=0 || cluster->ptx[i]!=0) && (cluster->pty[i]!=cluster->cly[n] || cluster->ptx[i]!=cluster->clx[n])){
		    f=(cluster->cly[n]-cluster->pty[i])*(cluster->cly[n]-cluster->pty[i])+(cluster->clx[n]-cluster->ptx[i])*(cluster->clx[n]-cluster->ptx[i]);
		    Qtan=atan2((cluster->cly[n]-cluster->pty[i])*1.0,(cluster->ptx[i]-cluster->clx[n])*1.0);
		    if(Qtan>=0.0){
			qq=(int)(0.5+Qtan*180/PI);
			if(qq==DEG) qq=0;
		    }
		    else qq=(int)(0.5+(Qtan+PI*2)*180/PI);
		    if(q[qq]<f) q[qq]=f;
		}
	    }
	}

	for(i=0;i<=DEG;i++){
	    if(q[i]<1){
		if(i<180){
		    for(j=i;j<=DEG;j++){
			if(q[j]>=1){
			    q[i]=q[j];
			    break;
			}
		    }
		}
		else{
		    for(j=i;j>0;j--){
			if(q[j]>=1){
			    q[i]=q[j];
			    break;
			}
		    }
		}
	    }
	}

	double mm[9];
	for(i=0;i<=DEG;i++){
	    k=0;
	    for(j=-4+i;j<=4+i;j++){
		if(j<0)
		    mm[k]=q[DEG+j];
		else if(j>=DEG)
		    mm[k]=q[j-DEG];
		else
		    mm[k]=q[j];
		k++;
	    }
	    F[i]=median(mm);
	}
	for(i=0;i<=DEG;i++){
	    for(j=-10+i;j<10+i;j++){
		if(j<0)
		    F[i]=F[i]+q[DEG+j];
		else if(j>=DEG)
		    F[i]=F[i]+q[j-DEG];
		else
		    F[i]=F[i]+q[j];
	    }
	    F[i]=F[i]/20;
	}


	for(i=0;i<BIT8;i++){
	    for(j=0;j<BIT8;j++){
		if(i==cluster->cly[n] && j==cluster->clx[n])
		    gaussian[i][j]=1;
		else{
		    dis=(cluster->cly[n]-i)*(cluster->cly[n]-i)+(cluster->clx[n]-j)*(cluster->clx[n]-j);
		    Qtan=atan2((cluster->cly[n]-i)*1.0,(j-cluster->clx[n])*1.0);
		    if(Qtan>=0.0){
			shita=(int)(0.5+Qtan*180/PI);
			if(shita==DEG) shita=0;
		    }
		    else shita=(int)(0.5+(Qtan+PI*2)*180/PI);
		    fmax=F[shita];

		    if(fmax!=0)
		      gaussian[i][j]=gaussian[i][j]+exp(-(dis)/(2*fmax));
		    //gaussian[i][j]=gaussian[i][j]+exp(-(dis)/(2*fmax*fmax));
		    else
			gaussian[i][j]=gaussian[i][j]+exp(-(dis)/2);
		}
	    }
	}
	for(i=0;i<BIT8;i++){
	    for(j=0;j<BIT8;j++){
		gaussian[i][j]=gaussian[i][j];
	    }
	}
	for(i=0;i<=DEG;i++){
	    q[i]=F[i]=0;
	}
    }
    max=0;
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    prehist[i][j]=prehist[i][j]+histgram[i][j]*gaussian[i][j];
	    if(prehist[i][j]>max)  max=prehist[i][j];
	}
    }
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    histgram[i][j]=Nf*prehist[i][j]/max;
	}
    }
}


void makeHist(IplImage* ext, double hist[][BIT8], HScolor hs[][BIT8][BIT8])
{
    int i,j;
    unsigned char *p; 
    double h[BIT8][BIT8]={{}};
    p=(unsigned char*)ext->imageData;

    for(i=0;i<ext->width*ext->height;i++){
	if(hs[p[i*3+2]][p[i*3+1]][p[i*3]].S!=0)
	    h[hs[p[i*3+2]][p[i*3+1]][p[i*3]].S][hs[p[i*3+2]][p[i*3+1]][p[i*3]].H]=h[hs[p[i*3+2]][p[i*3+1]][p[i*3]].S][hs[p[i*3+2]][p[i*3+1]][p[i*3]].H]+1;
    }
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    hist[i][j]=h[i][j];
	}
    }
}

void upHist(double histgram[][BIT8], double newhist[][BIT8], HShist hshist[], int count)
{
    int i;
    double max=0;
    double prehist[BIT8][BIT8]={{}};
    for(i=0;i<count;i++){
	prehist[hshist[i].y][hshist[i].x]=histgram[hshist[i].y][hshist[i].x]+newhist[hshist[i].y][hshist[i].x]*hshist[i].val;
	if(max<prehist[hshist[i].y][hshist[i].x])
	    max=prehist[hshist[i].y][hshist[i].x];
    }
    for(i=0;i<count;i++){
	histgram[hshist[i].y][hshist[i].x]=255*prehist[hshist[i].y][hshist[i].x]/max;
    }
}


void Crop(IplImage* src,IplImage* img, CvPoint pt1, CvPoint pt2)
{
    int i;
    char *t=img->imageData;
    char *s=src->imageData+pt1.x*3+(pt1.y*WIDTH)*3;
    int width=(pt2.x-pt1.x)*3;
    for(i=pt1.y;i<pt2.y;i++){
	memcpy(t,s,width);
	t+=width;
	s+=src->width*3;
    }
}

void NNClustering(double clusterhist[][BIT8], CLdata* cluster, int threshold)
{
    int i,j;
    double distance,min;
    int min_num=0;
    cluster->pt=0;
    for(i=0;i<BIT8;i++){
	for(j=0;j<BIT8;j++){
	    if(clusterhist[i][j]>0){
		cluster->ptx[cluster->pt]=j;  
		cluster->pty[cluster->pt]=i;
		cluster->pt++;
	    }
	}
    } 

    cluster->clpt=0;
    cluster->clx[cluster->clpt]=cluster->ptx[cluster->clpt];
    cluster->cly[cluster->clpt]=cluster->pty[cluster->clpt];
    cluster->clnum[cluster->clpt]=0;
    cluster->clpt++;

    for(i=0;i<cluster->pt;i++){
	min=10000;
	for(j=0;j<cluster->clpt;j++){
	    distance=sqrt(((cluster->ptx[i] - cluster->clx[j])*(cluster->ptx[i] - cluster->clx[j])+(cluster->pty[i] - cluster->cly[j])*(cluster->pty[i] - cluster->cly[j]))*1.0);
	    if(distance<min){
		min=distance;
		min_num=j;
	    }
	}
	if(min<threshold){
	    cluster->clnum[i]=min_num;
	    cluster->clx[min_num]+=cluster->ptx[i];
	    cluster->clx[min_num]/=2;
	    cluster->cly[min_num]+=cluster->pty[i];
	    cluster->cly[min_num]/=2;
	}
	else{
	    cluster->clx[cluster->clpt]=cluster->ptx[i];
	    cluster->cly[cluster->clpt]=cluster->pty[i];
	    cluster->clnum[i]=cluster->clpt;
	    cluster->clpt++;
	}
    }

}

void MedianFilter(double hist[][BIT8])
{ 
    int i,j;
    double med[9];
    for(i=0;i<BIT8-1;i++){
	for(j=0;j<BIT8-1;j++){
	    med[0]=hist[i+1][j+1];
	    med[1]=hist[i+1][j-1];
	    med[2]=hist[i+1][j  ];
	    med[3]=hist[i-1][j+1];
	    med[4]=hist[i-1][j-1];
	    med[5]=hist[i-1][j  ];
	    med[6]=hist[i  ][j+1];
	    med[7]=hist[i  ][j-1];
	    med[8]=hist[i  ][j  ];
	    hist[i][j]=median(med);
	}
    }
}

double median(double m[])
{
    int i,j;
    double ori;
    for(j=0;j<8;j++){
	for(i=0;i<8;i++){
	    if(m[i+1]<m[i]){
		ori=m[i+1];
		m[i+1]=m[i];
		m[i]=ori;
	    }
	}
    }
    return m[4];
}



void inputHShist(double gaussian[][BIT8],HShist hsGaussian[],int &plot_count)
{
    int k=0;
    for(int i=0;i<BIT8;i++){
	for(int j=0;j<BIT8;j++){
	    if(gaussian[i][j]>0){
		hsGaussian[k].x=j;
		hsGaussian[k].y=i;
		hsGaussian[k].val=gaussian[i][j];
		k++;
	    }
	}
    }

    plot_count=k;
}



void NoiseFilter(double histgram[][BIT8]){
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			if(histgram[i][j]<2)
				histgram[i][j]=0;
		}
	}
}

void RectDetec(double histgram[][BIT8],HScolor hstable[][BIT8][BIT8],IplImage *img,int *_proxmax,int *_proymax,CvPoint *_xy,CvPoint *_xy1){
	int k=0;
	int j=0;
	int i;
	CvPoint xy,xy1;
	xy.x=0;
	xy.y=0;
	xy1.x=0;
	xy1.y=0;
	unsigned char *data=(unsigned char*)img->imageData;
	int proxmax=0;
	int proymax=0;
	int promax_x=0;
	int promax_y=0;
	int prox[WIDTH]={};
	int proy[HEIGHT]={};

	for(int i=0;i<WIDTH;i++)prox[i]=0;
	for(int i=0;i<HEIGHT;i++)proy[i]=0;

	for(int i=0;i<WIDTH*HEIGHT;i++){
		if(i>14){
	if((data[i*3]!=0)&&(data[i*3+1]!=0)&&(data[i*3+2]!=0)){
			if(histgram[hstable[data[i*3+2]][data[i*3+1]][data[i*3]].S][hstable[data[i*3+2]][data[i*3+1]][data[i*3]].H]>1){
				prox[j]=prox[j]+1;
				proy[k]=proy[k]+1;
			}
		}
}
		j++;
		if(j==WIDTH){
			j=0;
			k++;
		}
	}
	prox[0]=prox[WIDTH-1]=proy[0]=proy[HEIGHT-1]=proxmax=proymax=promax_x=promax_y=0;

	for(i=0;i<WIDTH;i++){
		if(prox[i]<5)  prox[i]=0;
		if(prox[i]>proxmax) { proxmax=prox[i]; promax_x=i; }
	}

	for(i=0;i<HEIGHT;i++){
		if(proy[i]<5)  proy[i]=0;
		if(proy[i]>proymax) { proymax=proy[i]; promax_y=i; }
	}
	for(i=promax_x;i>0;i--)   { if(prox[i]<=20) { xy.x=i; break; } } 
	for(i=promax_x;i<WIDTH;i++) { if(prox[i]<=20) { xy1.x=i; break; } }
	for(i=promax_y;i>0;i--)   { if(proy[i]<=10) { xy.y=i; break; } } 
	for(i=promax_y;i<HEIGHT;i++) { if(proy[i]<=10) { xy1.y=i; break; } }

	_xy->x=xy.x;
	_xy->y=xy.y;
	_xy1->x=xy1.x;
	_xy1->y=xy1.y;
	*_proxmax=proxmax;
	*_proymax=proymax;
}


void OutputHsv(double histgram[][BIT8],IplImage* hsv){
    int k=0;
	double max_hist=0;
	double hist=0;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			hist=(double)histgram[i][j];
			if(hist > max_hist) max_hist=hist;			
			k++;
	}
    }
	k=0;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			hsv->imageData[k]=(char)(255-histgram[i][j]);
			k++;
		}
	}
}

void OutputExt(double histgram[][BIT8],IplImage* hsg){
    int k=0;
	double max_hist=0;
	double hist=0;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			hist=(double)histgram[i][j];
			if(hist > max_hist) max_hist=hist;			
			k++;
	}
    }

	k=0;
	//cout<<max_hist<<endl;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			hsg->imageData[k]=(char)(255.0*1.0-255.0*histgram[i][j]/max_hist);
			k++;
	}
    }
}
void OutputGau(double gaussian[][BIT8],IplImage* gau){

	int k=0;
	double max_hist=0;
	double hist=0;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			hist=(double)gaussian[i][j];
			if(hist > max_hist) max_hist=hist;			
			k++;
		}
	}
//cout<<max_hist<<endl;
	k=0;
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			if(gaussian[i][j]>0) gau->imageData[k]=(char)(255-255*gaussian[i][j]/max_hist);
			else gau->imageData[k]=(char)255;
			k++;
		}
	}

}

void ExtHistInit(double ExtHist[][BIT8]){
    for(int i=0;i<BIT8;i++){
	for(int j=0;j<BIT8;j++){
	    ExtHist[i][j]=0;
	}
    }
}

void initHist(double hist[][BIT8]){
    for(int i=0;i<BIT8;i++){
	for(int j=0;j<BIT8;j++){
	    hist[i][j]=0;
	}
    }
}

void initGau(double gau[][BIT8]){
	for(int i=0;i<BIT8;i++){
		for(int j=0;j<BIT8;j++){
			gau[i][j]=0;
		}
	} 
}


void ColorExtract2(IplImage* target,IplImage *img,double histgram[][BIT8],HScolor hstable[][BIT8][BIT8]){
    //uchar b,g,r;
    unsigned char* data=(unsigned char*)img->imageData;

    for (int y = 0; y < HEIGHT; y++) {
	for (int x = 0; x < WIDTH; x++) {
	    int i=WIDTH*y+x;
	    target->imageData[target->widthStep * y + x * 3] = (char)0;
	    target->imageData[target->widthStep * y + x * 3 + 1] = (char)0;
	    target->imageData[target->widthStep * y + x * 3 + 2] =(char)0;
	    if((data[i*3]!=0)&&(data[i*3+1]!=0)&&(data[i*3+2]!=0)){
	    if((data[i*3]!=255)&&(data[i*3+1]!=255)&&(data[i*3+2]!=255)){
		if(histgram[hstable[data[i*3+2]][data[i*3+1]][data[i*3]].S][hstable[data[i*3+2]][data[i*3+1]][data[i*3]].H]>1){
		    target->imageData[target->widthStep * y + x * 3] = data[i*3];
		    target->imageData[target->widthStep * y + x * 3 + 1] = data[i*3+1];
		    target->imageData[target->widthStep * y + x * 3 + 2] =data[i*3+2];
		}
}
	    }
	}
    }
}


